# Generated from java-escape by ANTLR 4.5
from antlr4 import *

# This class defines a complete listener for a parse tree produced by PLambdaParser.
class PLambdaListener(ParseTreeListener):

    # Enter a parse tree produced by PLambdaParser#unit.
    def enterUnit(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#unit.
    def exitUnit(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#seqExpression.
    def enterSeqExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#seqExpression.
    def exitSeqExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#letExpression.
    def enterLetExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#letExpression.
    def exitLetExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#defineExpression.
    def enterDefineExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#defineExpression.
    def exitDefineExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#lambdaExpression.
    def enterLambdaExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#lambdaExpression.
    def exitLambdaExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#invokeExpression.
    def enterInvokeExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#invokeExpression.
    def exitInvokeExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#applyExpression.
    def enterApplyExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#applyExpression.
    def exitApplyExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#dataExpression.
    def enterDataExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#dataExpression.
    def exitDataExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#unaryExpression.
    def enterUnaryExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#unaryExpression.
    def exitUnaryExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#binaryExpression.
    def enterBinaryExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#binaryExpression.
    def exitBinaryExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#ternaryExpression.
    def enterTernaryExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#ternaryExpression.
    def exitTernaryExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#oneOrMoreExpression.
    def enterOneOrMoreExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#oneOrMoreExpression.
    def exitOneOrMoreExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#twoOrMoreExpression.
    def enterTwoOrMoreExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#twoOrMoreExpression.
    def exitTwoOrMoreExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#naryExpression.
    def enterNaryExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#naryExpression.
    def exitNaryExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#tryExpression.
    def enterTryExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#tryExpression.
    def exitTryExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#forExpression.
    def enterForExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#forExpression.
    def exitForExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#stringLiteral.
    def enterStringLiteral(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#stringLiteral.
    def exitStringLiteral(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#identifierLiteral.
    def enterIdentifierLiteral(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#identifierLiteral.
    def exitIdentifierLiteral(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#noneLiteral.
    def enterNoneLiteral(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#noneLiteral.
    def exitNoneLiteral(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#parameterList.
    def enterParameterList(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#parameterList.
    def exitParameterList(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#parameter.
    def enterParameter(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#parameter.
    def exitParameter(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#bindingList.
    def enterBindingList(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#bindingList.
    def exitBindingList(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#bindingPair.
    def enterBindingPair(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#bindingPair.
    def exitBindingPair(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#catchExpression.
    def enterCatchExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#catchExpression.
    def exitCatchExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#rangeExpression.
    def enterRangeExpression(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#rangeExpression.
    def exitRangeExpression(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#data.
    def enterData(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#data.
    def exitData(self, ctx):
        pass


    # Enter a parse tree produced by PLambdaParser#token.
    def enterToken(self, ctx):
        pass

    # Exit a parse tree produced by PLambdaParser#token.
    def exitToken(self, ctx):
        pass





import sys

def isString(data):
    if (sys.version_info[0] == 3):
        return isinstance(data, str)
    return isinstance(data, basestring)


def isInteger(data):
    if (sys.version_info[0] == 3):
        return isinstance(data, int)
    return isinstance(data, (int, long))



def string2File(string, path, append):
    retval = False
    try:
        mode = 'a' if append else 'w'
        with open(path, mode) as fp:
            fp.write(string)
        retval = True
    except Exception as e:
        sys.stderr.write('string2File(..., {0}, ...) threw {1}\n'.format(path, str(e)))
    return retval


def file2String(path):
    retval = None
    try:
        with open(path, 'r') as fp:
            retval = fp.read()
    except Exception as e:
        sys.stderr.write('file2String("{0}") threw {1}\n'.format(path, str(e)))
    return retval



import psutil, os, re, sys, signal, subprocess, time, threading, traceback

from .actorlib import send, receive

from ..eval.Interpreter import Interpreter

from ..visitor.Parser import parseFromString

from ..actors.console import Console

debug = False

def infanticide(pid):
    try:
      parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
      return
    children = parent.children(recursive=True)
    if debug:
        sys.stderr.write('The children of {0} are {1}\n'.format(pid, children))
    for p in children:
        os.kill(p.pid, signal.SIGKILL)
        if debug:
            sys.stderr.write('Sent signal {1} to {0}\n'.format(p.pid, signal.SIGKILL))


def sighandler(signum, frame):
    if debug:
        sys.stderr.write('Signal handler called with signal {0}\n'.format(signum))
        sys.stderr.flush()
    infanticide(os.getpid())
    sys.exit()


def main():
    if len(sys.argv) == 1:
        launch("noname")
    elif len(sys.argv) == 2:
        launch(sys.argv[1])
    else:
        launch(sys.argv[1], sys.argv[2])
    return 0


class Main(object):

    retries = 1

    launchConsole = False

    #iam: must be a better way
    plambda_pattern = re.compile('^\s*\(')

    handlers = []

    # a singleton instance
    myself = None
    
    def __init__(self, name, filename=None):
        """Creates an plambda Actor object with the given name.
        """
        self.name = name
        self.interpreter = Interpreter()
        self.filename = filename
        signal.signal(signal.SIGINT, sighandler)
        if Main.myself is None:
            Main.myself = self
        else:
            raise Exception("plambda.actor.pyactor.Main should have a singleton instance!")
        if self.filename is not None:
            notify('Loading {0}\n'.format(self.filename))
            self.interpreter.load(self.filename)


        
    def run(self):
        """The main thread ever ready to launch the console.
        """
        fails = 0

        self.oracle = threading.Thread(target=oracle, args=(self, ))
        self.oracle.daemon = True
        self.oracle.start()
        
        while True:
            time.sleep(1)
            #FIXME: better do this with a message now that we have handlers.
            if Main.launchConsole:
                Main.launchConsole = False
                self.console = Console(self.interpreter)
                self.console.mainloop()

                
        return 0


def add_handler(closure):
    Main.handlers.append(closure)
    
def remove_handler(closure):
    Main.handlers.remove(closure)
    
    
def notify(message):
    if debug:
        sys.stderr.write('{0}\n'.format(message))
        sys.stderr.flush()


def oracle(actor):
    """The Read Eval Message Loop.
    """
    fails = 0

    while True:
        incoming = receive()
        if incoming is None:
            fails += 1
            if fails > Main.retries:
                return 0
            else:
                continue
        (sender, msg) = incoming
        thread = threading.Thread(target=eval, args=(actor.interpreter, sender, msg))
        #thread needs to be a daemon so that the actor itself can die in peace.
        thread.daemon = True
        thread.start()
            

        
def eval(interpreter, sender, message):
    """Evaluates the message in the given interpreter.

    If the message appears to be an PLambda expression it is simply
    dispatched to the interpreter to evaluate. Otherwise it is processed
    by any custom handlers that have been installed.

    Custom handlers are of the form:
    
    (define handler (lambda (sender message)  ... ))

    and are installed by 

    (import 'plambda.actors.pyactor')
    (apply plambda.actors.pyactor.add_handler handler)


    and removed analgously


    (apply plambda.actors.pyactor.remove_handler handler)



    """
    try:
        if Main.plambda_pattern.match(message):
            val = interpreter.evaluateString(message)
            notify('eval: {0} evaluated to {1}'.format(message, val))
        else:
            sender = sender.strip()
            message = message.strip()
            notify('looking for handlers for: "{0}" from {1}'.format(message, sender))
            for handler in Main.handlers:
                if handler.applyClosure(sender, message):
                    notify('found a handler for: "{0}" from sender {1}'.format(message, sender))
                    break
            
    except Exception as e:
        sys.stderr.write('plambda.actors.pyactor.Main exception: {0}\n'.format(e))
        if debug:
            traceback.print_exc(file=sys.stderr)


def launch(name, file=None):
    main = Main(name, file)
    main.run()
    return 0
    



import io

class StringBuffer(object):

  def __init__(self):
    self.empty = True;
    self._stringio = io.StringIO()
    
  def __str__(self):
    val = self._stringio.getvalue()
    self._stringio.close()
    return val

  def append(self, obj):
    data = unicode(obj)
    if self.empty and len(data) > 0:
      self.empty = False
    self._stringio.write(unicode(obj))
    return self

  def isempty(self):
    return self.empty





"""
Freeping Creaturism:

This is the all important version number used by pip.


Version history:

1.0.0   7/1/2016    Paging in the pip process.

1.0.1   8/5/2016    pyactor shutsdown (i.e. kills) any children (like SITL etc).

1.0.2   8/18/2016   kwapply added.

1.0.3   8/18/2016   modify added.

1.0.4   8/21/2016   single quote strings added. isint isfloat added. isobject just (not (isnone))

1.0.5   8/23/2016   The read-eval-print loop gets a makeover. handles multiline paste and prints definitions.

1.0.6   8/25/2016   PLambda gets a console.

1.0.7   8/27/2016   Updated all relative imports to follow https://www.python.org/dev/peps/pep-0328/

1.0.8   9/02/2016   CPS (Continuation Passing Stype) installed as the default. Can switch to recursive via a mode in plambda.eval.Interpreter

1.0.9   9/10/2016   Various bug fixes in the CPS interpreter.

1.0.10  9/27/2016   Fixed the lookup mechanism in the CPS interpreter.

1.0.11  10/27/2016  Non Sexpression handlers added to pyactor for SoftAgent requirements.

"""

plambda_version='1.0.11'




from ..util.StringBuffer import StringBuffer

class PLambdaException(Exception):

    def __init__(self, msg, backtrace=None, exception=None):
        self.msg = msg
        self.backtrace = backtrace
        self.exception = exception
        
    def __str__(self):
        return str(self.msg)

    def __repr__(self):
        sb = StringBuffer()
        sb.append(self.msg)
        if self.backtrace:
            sb.append(' ').append(self.backtrace)
        if self.exception:
            sb.append(' ').append(self.exception)
        return str(sb)
        
    def extendBT(self, bt):
        self.bt = '{0}\n{1}'.format(self.backtrace, bt)



from .Code import SExpression, Atom, StringLiteral, Syntax

from .Closure import Closure

from .SymbolTable import SymbolTable

from .Continuation import ( TopCont, IfCont, SeqCont, DefineCont, TryCont, LetCont, ForCont,
                            UnaryOpCont, BinaryOpCont, TernaryOpCont, Ambi1OpCont,
                            ConcatCont, MkCont,
                            AndCont, OrCont,
                            ApplyCont, InvokeCont, GetAttrCont )


EVAL     = 0
CONTINUE = 1
RETURN   = 2
DONE     = 3



class State(object):

    def __init__(self, interpreter, exp, env):
        self.interpreter = interpreter
        self.tag = EVAL
        self.val = None
        self.exp = exp
        self.env = env
        self.k = TopCont()


    def isDone(self):
        return self.tag is DONE


    def step(self):

        if self.tag is EVAL:
            if  self.exp is None:
                self.val = None
                self.tag = RETURN
            elif isinstance(self.exp, StringLiteral):
                self.val = self.exp.string
                self.tag = RETURN
            elif isinstance(self.exp, Atom):
                (ok, value) = self.interpreter.lookup(self.exp, self.env)
                if ok:
                    self.val = value
                    self.tag = RETURN
                else:
                    self.k.excep = value
                    self.tag = RETURN
            elif isinstance(self.exp, SExpression):
                code = self.exp.code
                opexp = self.exp.spine[0]
                if not isinstance(opexp, Atom):
                    print 'not atom: {0} {1} {2} {3}'.format(opexp, type(opexp), self.exp, self.exp.code)
                assert isinstance(opexp, Atom)
                op = opexp.string
                if code is Syntax.SEQ:
                    self.k = SeqCont(self.exp, self.exp.spine[1:], self.env, self.k);
                    self.tag = CONTINUE
                elif code is Syntax.LET:
                    self.k = LetCont(self.exp, self.exp.spine[1:], self.env, self.k);
                    self.tag = CONTINUE
                elif code is Syntax.DEFINE:
                    self.k = DefineCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.LAMBDA:
                    spine = self.exp.spine
                    self.val = Closure(self.interpreter, spine[1], spine[2], self.env, spine[0].location)
                    self.tag = RETURN
                elif code is Syntax.INVOKE:                    
                    self.k = InvokeCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.APPLY:
                    self.k = ApplyCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.PRIMITIVE_DATA_OP:
                    self.val = self.interpreter.evalPrimitiveDataOp(self.exp, self.env)
                    self.tag = RETURN
                elif code is Syntax.UNARY_OP:
                    self.k = UnaryOpCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.BINARY_OP:                    
                    self.k = BinaryOpCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.TERNARY_OP:
                    self.k = TernaryOpCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.AMBI1_OP:
                    self.k = Ambi1OpCont(op, self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.AMBI2_OP:
                    if op is SymbolTable.IF:
                        self.k = IfCont(self.exp, self.exp.spine[1:], self.env, self.k)
                        self.tag = CONTINUE
                    elif op is SymbolTable.GETATTR:
                        self.k = GetAttrCont(self.exp, self.exp.spine[1:], self.env, self.k)
                        self.tag = CONTINUE
                    else:
                        raise PLambdaException("Unhandled ambi2 op form in State.step {0} {1}".format(op, self.exp.spine[0].location))
                elif code is Syntax.N_ARY_OP:
                    if op is SymbolTable.AND:
                        self.k = AndCont(self.exp, self.exp.spine[1:], self.env, self.k)
                        self.tag = CONTINUE
                    elif op is SymbolTable.OR:
                        self.k = OrCont(self.exp, self.exp.spine[1:], self.env, self.k)
                        self.tag = CONTINUE
                    elif op is SymbolTable.CONCAT:
                        self.k = ConcatCont(self.exp, self.exp.spine[1:], self.env, self.k)
                        self.tag = CONTINUE
                    elif op in (SymbolTable.MKTUPLE, SymbolTable.MKLIST, SymbolTable.MKDICT):
                        self.k = MkCont(op, self.exp, self.exp.spine[1:], self.env, self.k)
                        self.tag = CONTINUE
                    else:
                        raise PLambdaException("Unhandled n-ary op form in State.step {0} {1}".format(op, self.exp.spine[0].location))
                elif code is Syntax.TRY:
                    self.k = TryCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.FOR:
                    self.k = ForCont(self.exp, self.exp.spine[1:], self.env, self.k)
                    self.tag = CONTINUE
                elif code is Syntax.CATCH:
                    raise PLambdaException("Orphan catch. Should not happen")
                else:
                   raise PLambdaException("Unhandled form in State.step")
            else:
                self.val = self.exp
                self.tag = RETURN
        elif self.tag is CONTINUE:
            self.k.cont(self)
        elif self.tag is RETURN:
            self.k.ret(self)
        elif self.tag is DONE:
            pass
        else:
            pass





class Environment(object):

    def __init__(self, env = None):
        self.frame = {}
        self.next = env

    def extend(self, key, value):
        self.frame[key.string] = value
        return self
    
    def lookup(self, key):
        keystr = key.string
        env = self
        while env is not None:
            #print 'looking up {0} in {1}'.format(keystr, env.frame)
            if keystr in env.frame:
                return (True, env.frame[keystr])
            else:
                env = env.next
        return (False, None)
    
        
        
        

    



import sys

debug = False

def send(target, sender, msg):
    """Send a message to actor 'target' from actor 'sender' with content 'msg'.

    This uses iop's underlying message format.
    """
    
    bytes = '{0}\n{1}\n{2}'.format(target, sender, msg)
    length = len(bytes)
    outgoing = '{0}\n{1}'.format(len(bytes), bytes)
    if debug:
        sys.stderr.write(outgoing)
    sys.stdout.write(outgoing)
    sys.stdout.flush()

def parseBytes(bytes, nobytes):
    if bytes[-1] == '\n':
        bytes = bytes[:-1]
    if bytes[0] == '(' and bytes[-1] == ')':
            nl = bytes.find(' ')
            if nl > 0:
                return (True, bytes[1: nl], bytes[nl:-1])
    return (False, None, bytes)
    
def receive():
    """Reads an actor message and parses it as such. 

    Returns (sender, msg) if successful, 
    otherwise (None, bytes)
    """
    try:
        nobytes = sys.stdin.readline().strip()
        if debug:
            sys.stderr.write('got {0} bytes\n'.format(nobytes))
        bytes = sys.stdin.read(int(nobytes))
        if debug:
            sys.stderr.write('[[{0}]]\n'.format(bytes))
        (ok, sender, msg) = parseBytes(bytes, nobytes)
        if ok:
            return (sender, msg)
        else:
            return (None, bytes)
    except KeyboardInterrupt:
        return None
    except Exception as e:
        print e
        return (None, None)



    



import Tkinter as tk

import tkFileDialog

import os, sys, traceback

from ..eval.Interpreter import Interpreter

from ..visitor.Parser import parseFromString

from ..eval.PLambdaException import PLambdaException

from ..util.StringBuffer import StringBuffer


class InputTextArea(tk.Frame):
     def __init__(self, console):
         tk.Frame.__init__(self, borderwidth=1, relief="sunken")
         self.console = console
         self.text = tk.Text(wrap=tk.NONE, background="white", borderwidth=0, highlightthickness=0)
         self.text_sb = tk.Scrollbar(orient="vertical", borderwidth=1, command=self.text.yview)
         self.text.configure(yscrollcommand=self.text_sb.set)
         self.text_sb.pack(in_=self, side="right", fill="y", expand=False)
         self.text.pack(in_=self, side="left", fill="both", expand=True)
         self.text.bind("<Control-b>", lambda e: console.evaluate(self.buffer(), e))
         self.text.bind("<Command-b>", lambda e: console.evaluate(self.buffer(), e))
         self.text.bind("<Control-l>", lambda e: console.evaluate(self.line(), e))
         self.text.bind("<Command-l>", lambda e: console.evaluate(self.line(), e))
         self.text.bind("<Control-e>", lambda e: console.evaluate(self.selected(), e))
         self.text.bind("<Command-e>", lambda e: console.evaluate(self.selected(), e))
         self.text.bind("<Control-s>", lambda e: console.save())
         self.text.bind("<Command-s>", lambda e: console.save())
         self.text.bind("<Control-r>", lambda e: console.load(console.path))
         self.text.bind("<Command-r>", lambda e: console.load(console.path))


     def selected(self):
          if self.text.tag_ranges(tk.SEL):
               return self.text.get(tk.SEL_FIRST, tk.SEL_LAST)
          else:
               return None

     def line(self):
          return self.text.get("insert linestart", "insert lineend")

     def buffer(self):
          return self.text.get(1.0, tk.END)

          
class OutputTextArea(tk.Frame):
     def __init__(self):
         tk.Frame.__init__(self, borderwidth=1, relief="sunken")
         self.text = tk.Text(wrap=tk.NONE, background="white", borderwidth=0, highlightthickness=0, state=tk.DISABLED)
         self.text_sb = tk.Scrollbar(orient="vertical", borderwidth=1, command=self.text.yview)
         self.text.configure(yscrollcommand=self.text_sb.set)
         self.text_sb.pack(in_=self, side="right", fill="y", expand=False)
         self.text.pack(in_=self, side="left", fill="both", expand=True)

         self.text.tag_config("error", background="white", foreground="red")
         self.text.tag_config("ok", background="white", foreground="black")
         
     def append(self, text, tag):
          self.text.config(state=tk.NORMAL)
          self.text.insert(tk.END, text, tag)
          self.text.config(state=tk.DISABLED)
          self.text.see(tk.END)


class FileMenu(tk.Menu):
     def __init__(self, console):
          tk.Menu.__init__(self, console.menubar, tearoff=0)
          self.add_command(label="Open", command=console.open, accelerator="Command-O")
          self.add_command(label="Reload", command=lambda : console.load(console.path), accelerator="Command-R")
          self.add_command(label="Save", command=lambda : console.save(), accelerator="Command-S")
          self.add_command(label="Save As", command=console.saveas, accelerator="Shift-Command-W")
          self.add_command(label="Quit", command=sys.exit, accelerator="Command-Q")
          console.menubar.add_cascade(label="File", menu=self)

class EvaluateMenu(tk.Menu):
     def __init__(self, console):
          tk.Menu.__init__(self, console.menubar, tearoff=0)
          self.add_command(label="Buffer",
                           command=lambda : console.evaluate(console.top_frame.buffer(), None),
                           accelerator="Command-B")
          self.add_command(label="Selected",
                           command=lambda : console.evaluate(console.top_frame.selected(), None),
                           accelerator="Command-E")
          self.add_command(label="Line",
                           command=lambda : console.evaluate(console.top_frame.line(), None),
                           accelerator="Command-L")
          console.menubar.add_cascade(label="Evaluate", menu=self)

class ViewMenu(tk.Menu):
     def __init__(self, console):
          tk.Menu.__init__(self, console.menubar, tearoff=0)
          self.add_command(label="Definitions",
                           command=console.definitions,
                           accelerator="Command-D")
          self.add_command(label="UIDs",
                           command=console.uids,
                           accelerator="Command-U")
          self.add_command(label="Code",
                           command=console.code)
          console.menubar.add_cascade(label="View", menu=self)


class Console(tk.Tk):

     def __init__(self, interpreter):
        tk.Tk.__init__(self)
        self.title("PLambda Console")
        self.interpreter = interpreter
        self.console_frame = tk.Frame(borderwidth=1, relief="sunken")
        self.toolbar = tk.Frame()
        self.splitpane = tk.PanedWindow(orient=tk.VERTICAL, sashwidth=10, showhandle=True, sashrelief=tk.RAISED)
        self.splitpane.pack(fill=tk.BOTH, expand=1)
        self.top_frame = InputTextArea(self)
        self.splitpane.add(self.top_frame)
        self.bottom_frame = OutputTextArea()
        self.splitpane.add(self.bottom_frame)
        self.splitpane.pack(in_=self.console_frame, side="left", fill="both", expand=True)
        self.toolbar.pack(in_=self.console_frame, side="top", fill="x")
        self.menubar =  tk.Menu(self)
        self.filemenu =  FileMenu(self)
        self.evaluatemenu =  EvaluateMenu(self)
        self.viewmenu =  ViewMenu(self)
        self.config(menu=self.menubar)
        self.console_frame.pack(side="bottom", fill="both", expand=True)
        self.load("console.lsp")
        
     def evaluate(self, text, event):
          if self.interpreter is None:
               self.bottom_frame.append("Interpreter is None\n", "error")
          elif text is None:
               self.bottom_frame.append("Text to evaluate is None\n", "error")
          else:
               try:
                    code = parseFromString(text)
                    for c in code:
                         if c is not None:
                              value = self.interpreter.evaluate(c)
                              self.bottom_frame.append(str(value), "ok")
                              self.bottom_frame.append('\n', "ok")
               except PLambdaException as e:
                    self.bottom_frame.append(str(e), "error")
                    self.bottom_frame.append('\n', "ok")
               except Exception as e:
                    print 'PLambda.rep Exception: ', e
                    traceback.print_exc(file=sys.stderr)
                    self.bottom_frame.append('\n', "ok")

     def tbi(self):
          self.bottom_frame.append("To be implemented...\n", "error")

     def load(self, fname):
          if not fname:
               return
          path = os.path.abspath(fname)
          if os.path.exists(path):
               with open (path, "r") as fp:
                    self.top_frame.text.delete(1.0, tk.END)
                    for line  in fp.readlines():
                         self.top_frame.text.insert(tk.END, line)
               self.path = path
               self.bottom_frame.append("{0} loaded\n".format(path), "ok")
          else:
               self.bottom_frame.append("{0} not found\n".format(path), "error")
               
     def save(self, path=None):
          if path is None:
               path = self.path
          if path is None:
               self.bottom_frame.append("No file set, use 'Save as'\n", "error")
          else:
               with open(path, 'w') as fp:
                    fp.write(self.top_frame.buffer())
               self.bottom_frame.append("{0} saved\n".format(path), "ok")

     def open(self):
          path = tkFileDialog.askopenfilename(initialdir=os.getcwd())
          if path:
               self.load(path)

     def saveas(self):
          path = tkFileDialog.asksaveasfilename(initialdir=os.getcwd())
          if path:
               self.save(path)

     def uids(self):
          self.intdump(self.interpreter.showUIDs)

     def definitions(self):
          self.intdump(self.interpreter.showDefinitions)

     def code(self):
          self.intdump(self.interpreter.showCode)

     def intdump(self, func):
          self.bottom_frame.append(str(func(StringBuffer())), "ok")
          
        
def launch():
     console=Console(Interpreter())
     console.mainloop()
    
        
if __name__ == "__main__":
     launch()




# iam: This is needed to stop python from treating 'print' as something special; rather than 
# just a builtin. Wonder how much of this hackery is in our future?
#
from __future__ import print_function

import importlib, sys, types, inspect, collections, os

from ..util.Util import isString, isInteger

from ..util.StringBuffer import StringBuffer

from .Code import SExpression, Atom, StringLiteral, Syntax

from .SymbolTable import SymbolTable

from .PLambdaException import PLambdaException

from .Environment import Environment

from .Closure import Closure

from .Globals import pythonGlobals

from ..visitor.Parser import parseFromFile, parseFromString

from .State import State

"""
Current bugs:

Things to add:

patch the builtin problem as much as possible

ForCont mystery

"""

RECURSIVE = 0
CPS       = 1
"""
Change to RECURSIVE for recursive interpreter.
"""
CURRENT_MODE = CPS

class Interpreter(object):

    def __init__(self):
        self.definitions = {}
        self.code = {}
        self.modules = {}
        self.uid2object = {}
        self.object2uid = {}
        if CURRENT_MODE is RECURSIVE:
            self.evaluateString = self.recursive_evaluateString
            self.evaluate = self.recursive_evaluate
            self.eval = self.recursive_eval
        else:
            self.evaluateString = self.cps_evaluateString
            self.evaluate = self.cps_evaluate
            self.eval = self.cps_eval
            

    def recursive_evaluateString(self, string):
        """Parses and evaluates a string as plambda expression.
        """
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        code = parseFromString(string)
        retval = None
        for c in code:
            retval = self.evaluate(c)
        return retval
        
    def recursive_evaluate(self, exp):
        """Evaluates an Sexpression, StringLiteral, or Atom.
        """
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        return self.eval(exp, Environment())

    def recursive_eval(self, exp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        if exp is None:
            return None
        if isinstance(exp, StringLiteral):
            return exp.string
        if isinstance(exp, Atom):
            (ok, value) = self.lookup(exp, env)
            if ok:
                return value
            else:
                raise value
            #return self.lookup(exp, env)
        elif isinstance(exp, SExpression):
            return self.evalSExpression(exp, env)
        else:
            raise PLambdaException("huh? I did not grok {0}".format(exp))

    def cps_evaluateString(self, string):
        """Parses and evaluates a string as plambda expression.
        """
        code = parseFromString(string)
        retval = None
        for c in code:
            retval = self.cps_evaluate(c)
        return retval
        
    def cps_evaluate(self, exp):
        """Evaluates an Sexpression, StringLiteral, or Atom.
        """
        return self.cps_eval(exp, Environment())

    def cps_eval(self, exp, env):
        state = State(self, exp, env)
        while not state.isDone():
            state.step()
        #FIXME: we raise the uncaught exception. not sure why we need to differ from jlambda
        if state.k.excep is not None:
            raise state.k.excep
        else:
            return state.val



    def lookup(self, leaf, env):
        """See if the identifier is bound in the extended environment.

        First see if it referencing a thing in a module. Then, failing that
        look if it has a value in the current lexical environment. As a last
        resort see if it has a global definition, either in our global environment
        or in Python's. If it has a value in Python's global environment and that
        vaue is callable it retruns that. Otherwise it raises a PLambdaException.
        """

        assert(isinstance(leaf, Atom))

        name = leaf.string
        path = name.split('.')

        (ok, value) = self.mlookup(path, leaf)
        if ok:
            return (ok, value)
        (ok, value) = env.lookup(leaf)
        if ok:
            return (ok, value)
        (ok, value) = self.glookup(path, leaf)
        if ok:
            return (ok, value)
        (ok, value) = self.plookup(leaf)
        if ok:
            return (ok, value)
        return (False, PLambdaException('Unbound variable: {0}'.format(repr(leaf))))
    
        
    def getmodule(self, path):
        """Finds the longest prefix of path that is in the modules dictionary.

        If finds a prefix returns (mod, path_remainder), else (None, path).
        """
        index = len(path)
        remainder = []
        current = None
        mod = None
        while index != 0:
            current = '.'.join(path[0:index])
            mod = self.modules.get(current)
            if mod is not None:
                break
            index -= 1
            remainder.insert(0, path[index])
        if mod is not None:
            return (mod, remainder)
        else:
            return (None, path)

    def getobject(self, obj, path):
        """Follows the path from object.

        Follows the path down into the obj, which should be a class or module,
        and if what lies at the end is x returns (True, x), else, if the path
        makes no sense at some stage, returns (False, None).
        """
        if path == []:
            return (True, obj)
        if obj is None:
            return (False, None)
        elif inspect.ismodule(obj):
            d = obj.__dict__
            k = path[0]
            if k in d:
                return self.getobject(obj.__dict__.get(k), path[1:])
            else:
                return (False, None)
        else:
            if hasattr(obj, path[0]):
                nobj = getattr(obj, path[0])
            else:
                return (False, None)
            return self.getobject(nobj, path[1:])
        
        
    def mlookup(self, path, leaf):
        """Just a quick 'n dirty hack at this point.
        """
        assert(isinstance(leaf, Atom))

        (mod, remainder) = self.getmodule(path)

        return self.getobject(mod, remainder)

    def glookup(self, path, leaf):
        """Looks up the symbol in the PLambda definitions."""
        assert(isinstance(leaf, Atom))

        ok = False
        
        if len(path) > 1:
            key = path[0]
            val = self.definitions.get(key)
            if val is not None:
                (ok, val) = self.getobject(val, path[1:])
        else:
            key = leaf.string
            ok = key in self.definitions
            if ok:
                val = self.definitions[key]
        
        if ok:
            return (True, val)
        else:
            return (False, None)

    def plookup(self, leaf):
        """Looks up the symbol in the Python environment.

        This is a hack at present, as a means of looking up global functions.
        This is why we bail if the value is not callable.
        """
        assert(isinstance(leaf, Atom))
        
        try:
            val = eval(leaf.string)
            if callable(val):
                return (True, val)
        except NameError:
            pass
        return (False, None)

    def evalSeq(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        tail = sexp.spine[1:]
        retval = None
        for s in tail:
            retval = self.eval(s, env)
        return retval


    def callInvoke(self, obj, methodname, vals, location):
        
        methods = inspect.getmembers(obj, callable)

        if not methods:
            fmsg = 'Object not invokable: {0} {1}'
            emsg = fmsg.format(obj, location)
            return (False, PLambdaException(emsg))

        method = None
        
        if not isString(methodname):
            fmsg = 'Method name not a string: {0} {1}'
            emsg = fmsg.format(methodname, location)
            return (False, PLambdaException(fmsg))

        
        for (name, value) in methods:
            if name == methodname:
                method = value
                break

        if method is None:
            emsg = 'No such method: {0} {1}'.format(methodname, location)
            return (False, PLambdaException(emsg))
        

        # Ugliness under python's hood:
        # Cannot get the argspec of a builtin
        # http://stackoverflow.com/questions/3276635/how-to-get-the-number-of-args-of-a-built-in-function-in-python
        # http://stackoverflow.com/questions/990016/how-to-find-out-the-arity-of-a-method-in-python
        #
        # Can avoid using inspect for simlpe cases:
        #        def arity(obj, method):
        #          return getattr(obj.__class__, method).func_code.co_argcount - 1 # remove self
        #
        if type(method) !=  types.BuiltinFunctionType:
            argspec = inspect.getargspec(method)
            # if it is an object we have to *not* count 'self',
            # but if it is a class we need to pass all the args!
            offset = 0
            if not inspect.ismodule(obj) and not inspect.isclass(obj):
                offset = 1
                ndefaults = len(argspec.defaults)  if argspec.defaults else 0
                nargs = len(argspec.args) - offset
                nvals = len(vals)
                # my guess is that we will need to revisit this. what do we
                # do when some of the defaults are used but not others?
                if (nvals < nargs - ndefaults) or  (nargs < nvals):
                    fmsg = 'Arity of {0} args {1} does not match the argspec: {2}. defaults: {3}'
                    emsg = fmsg.format(methodname, vals, argspec.args[offset:], defaults)
                    return (False, PLambdaException(emsg))

        retval = None

        try:
            retval = method(*vals)
            return (True, retval)
        except Exception as e:
            return (False, PLambdaException('invoke {0} threw {1}'.format(location, str(e))))


    
    def evalInvoke(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        objexp = sexp.spine[1]
        obj = self.eval(objexp, env)

        methodexp = sexp.spine[2]
        methodname  = self.eval(methodexp, env)

        args = sexp.spine[3:]

        vals = []
        for a in args:
            vals.append(self.eval(a, env))

        (ok, retval) = self.callInvoke(obj, methodname, vals, sexp.spine[0].location)

        if ok:
            return retval
        else:
            raise retval

    def evalDefine(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        identifier = sexp.spine[1]
        
        assert(isinstance(identifier, Atom))
        
        val = None

        if len(sexp.spine) == 3:
            val = self.eval(sexp.spine[2], env)
        else:
            val = Closure(self, sexp.spine[2], sexp.spine[3], env, sexp.spine[0].location)

        idstr = identifier.string
        
        dot = idstr.find('.')

        if dot >= 0:
            sys.stderr.write('Bad idea to have a  "." in an identifier: {0}\n'.format(repr(identifier)))
            
        
        self.definitions[idstr] = val
        self.code[idstr] = sexp
        return identifier
    

    def evalLet(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        nenv = Environment(env)
        bindings = sexp.spine[1].spine
        body = sexp.spine[2]

        for bpair in bindings:
            var = bpair.spine[0]
            assert(isinstance(var, Atom))
            nenv.extend(var, self.eval(bpair.spine[1], nenv))

        return self.eval(body, nenv)


    def callClosure(self, fun, vals, location):
        """DEPRECATED VERSION.
        """
        retval = None
        assert(isinstance(fun, Closure))
        if len(vals) != fun.arity:
            fmsg = 'Arities at apply  {3} do not match: closure with arity {0} applied to args {1} of length {2}'
            emsg = fmsg.format(fun.arity, vals, len(vals), location)
            return (False, PLambdaException(emsg))
        else:
            try: 
                retval = fun.applyClosure(*vals)
                return (True, retval)
            except Exception as e:
                return (False, e)
        

    def callCallable(self, fun, vals, location):
        retval = None
        assert(callable(fun))
        try: 
            retval = fun(*vals)
            return (True, retval)
        except Exception as e:
            return (False, e)
            

    def callApply(self, fun, vals, location):
        """DEPRECATED VERSION.
        """
        retval = None
        if not isinstance(fun, Closure) and not callable(fun):
            fmsg = 'Cannot apply {0} which evaluated to {1}'
            emsg = fmsg.format(funexp, fun)
            return (False, PLambdaException(emsg))

        if isinstance(fun, Closure):
            return self.callClosure(fun, vals, location)
        else:
            return self.callCallable(fun, vals, location)


    def evalApply(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        funexp  = sexp.spine[1]
        argexps = sexp.spine[2:]

        fun = self.eval(funexp, env)

        vals = []
        for arg in argexps:
            vals.append(self.eval(arg, env))

        (ok, retval) = self.callApply(fun, vals,  sexp.spine[0].location)

        if ok:
            return retval
        else:
            raise retval


    def evalFor(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        identifier = sexp.spine[1]
        rangeexp = sexp.spine[2]
        body = sexp.spine[3]

        rangeval = self.eval(rangeexp, env)

        if isInteger(rangeval):
            rangeval = range(0, rangeval)  #FIXME: not very efficient for big integers
            
        if isinstance(rangeval, collections.Iterable):
            retval = None
            for v in rangeval:
                nenv = Environment(env)
                retval = self.eval(body, nenv.extend(identifier, v))
            return retval
        else:
            fmsg = 'Range is not iterable: {0} evaluated to {1}'
            emsg = fmsg.format(rangeexp, rangeval)
            raise PLambdaException(emsg)


    def evalTry(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        tryexp = sexp.spine[1]
        catchexp = sexp.spine[2]
        retval = None
        try:
            retval = self.eval(tryexp, env)
        except Exception as e:
            catchid = catchexp.spine[1]
            catchbody = catchexp.spine[2]
            catchenv = Environment(env)
            retval = self.eval(catchbody, catchenv.extend(catchid, e))
        return retval


    def callBinaryOp(self, op, val0, val1, location):
        retval = None
        try:
            if  op is SymbolTable.PLUS:
                retval =  val0 + val1
            elif op is SymbolTable.TIMES:
                retval = val0 * val1
            elif op is SymbolTable.DIVIDE:
                retval = val0 / val1
            elif op is SymbolTable.MODULO:
                retval = val0 % val1
            elif op is SymbolTable.GT:
                retval = val0 < val1
            elif op is SymbolTable.LT:
                retval = val0 > val1
            elif op is SymbolTable.GET:
                retval = self.evalGet(val0, val1, location)
            elif op is SymbolTable.GEQ:
                retval = val0 >= val1
            elif op is SymbolTable.LEQ:
                retval = val0 <= val1
            elif op is SymbolTable.EQUALS:
                retval = val0 == val1
            elif op is SymbolTable.IS:
                retval = val0 is val1
            elif op is SymbolTable.NEQ:
                retval = val0 != val1
            elif op is SymbolTable.SETUID:
                retval = self.setUID(val0, val1, location)
            else:
                fmsg = 'Unrecognized binary operation: {0} {1}'
                emsg = fmsg.format(op, location)
                return (False, PLambdaException(emsg))
        except Exception as e:
            return (False, PLambdaException('callTernaryOp {0} {1} threw {2}'.format(op, location, str(e))))
        return (True, retval)
    
    def evalBinaryOp(self, sexp, env):
        """Recursive version. Soon to be deprecated.
        """
        (uop, arg0, arg1) =  sexp.spine
        assert isinstance(uop, Atom)
        op = uop.string

        val0 = self.eval(arg0, env)
        val1 = self.eval(arg1, env)

        (ok, retval) = self.callBinaryOp(op, val0, val1, sexp.location)

        if ok:
            return retval
        else:
            raise retval
        
    def evalGet(self, val0, val1, location):
        if (isinstance(val0, list) or isinstance(val0, tuple))  and isinstance(val1, int):
            return val0[val1]
        elif isinstance(val0, dict):
            return val0.get(val1)
        else:
            fmsg = 'Bad args to \"get\": {0} {1} {2}'
            emsg = fmsg.format(val0, val1, location)
            raise PLambdaException(emsg)

    def unsetUID(self, val0):
        if val0 in self.object2uid:
            uid = self.object2uid[val0]
            self.object2uid.pop(val0)
            self.uid2object.pop(uid)
            return True
        else:
            return False
       

    def setUID(self, val0, val1, location):
        if val0 is None:
            raise PLambdaException('setuid {0}: first argument cannot be None.'.format(str(location)))                  
        elif val1 is None:
            return self.unsetUID(val0)
        elif not isString(val1):
            raise PLambdaException('setuid {0}: val1 not a string.'.format(str(location)))
        else:
            if val1 in self.uid2object or val0 in self.object2uid:
                raise PLambdaException('setuid {0}: redefiniton'.format(str(location)))
            else:
                self.object2uid[val0] = val1
                self.uid2object[val1] = val0
                return True

    def evalKWApply(self, val0, val1, val2, loc):
        if not callable(val0):
            raise PLambdaException('kwapply {0}: 1st argument not callable'.format(str(loc)))
        if not isinstance(val1, list):
            raise PLambdaException('kwapply {0}: 2nd argument not a list'.format(str(loc)))
        if not isinstance(val2, dict):
            raise PLambdaException('kwapply {0}: 3rd argument not a dict'.format(str(loc)))
        return  val0(*val1, **val2)


    def evalModify(self, val0, val1, val2, loc):
        #be brave until we need to back down...
        if isinstance(val0, dict):
            val0[val1] = val2 
        elif isinstance(val0, list):
            l0 = len(val0)
            if isinstance(val1, int) and -l0 < val1 and val1 < l0:
                val0[val1] = val2
            else:
                raise PLambdaException('modify {0}: bad index to modify of list'.format(str(sexp.location)))
        else:
            raise PLambdaException('modify {0}: unhandled case'.format(str(sexp.location)))
        return None




    def callTernaryOp(self, op, val0, val1, val2, location):
        retval = None
        try:
            if  op in (SymbolTable.UPDATE, SymbolTable.SUPDATE, SymbolTable.SETATTR):
                setattr(val0, val1, val2)
                retval = None
            elif op is SymbolTable.KWAPPLY:
                retval = self.evalKWApply(val0, val1, val2, location)
            elif op is SymbolTable.MODIFY:
                retval = self.evalModify(val0, val1, val2, location)
            else:
                fmsg = 'Unrecognized ternary operation: {0} {1}'
                emsg = fmsg.format(op, location)
                return (False, PLambdaException(emsg))
        except Exception as e:
            return (False, PLambdaException('callTernaryOp {0} {1} threw {2}'.format(op, location, str(e))))
        return (True, retval)
        

    def evalTernaryOp(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        (uop, arg0, arg1, arg2) =  sexp.spine
        assert isinstance(uop, Atom)
        op = uop.string

        val0 = self.eval(arg0, env)
        val1 = self.eval(arg1, env)
        val2 = self.eval(arg2, env)

        (ok, retval) = self.callTernaryOp(op, val0, val1, val2, sexp.location)
        
        if ok:
            return retval
        else:
            raise retval
        
    def evalAmbi1Op(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        uop =  sexp.spine[0]
        assert isinstance(uop, Atom)
        op = uop.string
        if op is SymbolTable.MINUS:
            if len(sexp.spine) == 3:
                return self.eval(sexp.spine[1], env) - self.eval(sexp.spine[2], env)
            else:
                return - self.eval(sexp.spine[1], env)
        else:
            fmsg = 'Unrecognized ambi1 operation: {0}'
            emsg = fmsg.format(op)
            raise PLambdaException(emsg)


    def callGetAttr(self, vals, location):
        retval = None
        try:
            retval = getattr(*vals)
            return (True, retval)
        except Exception as e:
            return (False, PLambdaException('callGetAttr {0} threw {1}'.format(location, str(e))))
        
        
    def evalAmbi2Op(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        uop =  sexp.spine[0]
        args = sexp.spine[1:]
        assert isinstance(uop, Atom)
        op = uop.string
        if op is SymbolTable.IF:
            return self.evalIf(sexp, env)
        elif op in (SymbolTable.GETATTR, SymbolTable.LOOKUP):
            vals = []
            
            for arg in args:
                vals.append(self.eval(arg, env))
            (ok, retval) = self.callGetAttr(vals, sexp.spine[0].location)
            if ok:
                return retval
            else:
                raise retval
        else:
            fmsg = 'Unrecognized ambi1 operation: {0}'
            emsg = fmsg.format(op)
            raise PLambdaException(emsg)

    def evalIf(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        testexp = sexp.spine[1]
        test = self.eval(testexp, env)
        #do we want to enforce booleaness here? JLambda does.
        if not isinstance(test, bool):
            msg = '{0} is not a boolean in conditional {1}'
            raise PLambdaException(msg.format(test, sexp.spine[0].location))
        elif test:
            return  self.eval(sexp.spine[2], env)
        if len(sexp.spine) == 4:
            return  self.eval(sexp.spine[3], env)
        else:
            return None


    def evalAnd(self, args, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        for e in args:
            if  self.eval(e, env) is False:
                return False
        return  True
        

    def evalOr(self, args, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        for e in args:
            if  self.eval(e, env):
                return True
        return  False

        
    def evalNaryOp(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        uop =  sexp.spine[0]
        assert isinstance(uop, Atom)
        op = uop.string
        args = sexp.spine[1:]
        
        if op is SymbolTable.AND:
            return self.evalAnd(args, env)
        elif op is SymbolTable.OR:
            return self.evalOr(args, env)
        elif  op is SymbolTable.CONCAT:
            return  self.evalConcat(sexp, env)
        elif op in (SymbolTable.MKTUPLE, SymbolTable.MKLIST, SymbolTable.MKDICT):
            vals = []
            for arg in args:
                vals.append(self.eval(arg, env))
            if op is SymbolTable.MKTUPLE:
                return tuple(vals)
            elif op is SymbolTable.MKLIST:
                return vals
            else:
                return dict(vals[i:i+2] for i in range(0, len(vals), 2))
        else:
            fmsg = 'Unrecognized n-ary operation {1}: {0}'
            emsg = fmsg.format(op, sexp.location)
            raise PLambdaException(emsg)

    def evalConcat(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        args = sexp.spine[1:]
        sb = StringBuffer()
        for e in args:
            sb.append(str(self.eval(e, env)))
        return  str(sb)
                               
    def evalSExpression(self, sexp, env):
        """RECURSIVE VERSION. Soon to be deprecated.
        """
        code = sexp.code
        if code is Syntax.SEQ:
            return self.evalSeq(sexp, env)
        elif code is Syntax.LET:
            return self.evalLet(sexp, env)
        elif code is Syntax.DEFINE:
            return self.evalDefine(sexp, env)
        elif code is Syntax.LAMBDA:
            spine = sexp.spine
            return Closure(self, spine[1], spine[2], env, spine[0].location)
        elif code is Syntax.INVOKE:
            return self.evalInvoke(sexp, env)
        elif code is Syntax.APPLY:
            return self.evalApply(sexp, env)
        elif code is Syntax.PRIMITIVE_DATA_OP :
            return self.evalPrimitiveDataOp(sexp, env)
        elif code is Syntax.UNARY_OP:
            return self.evalUnaryOp(sexp, env)
        elif code is Syntax.BINARY_OP:
             return self.evalBinaryOp(sexp, env)
        elif code is Syntax.TERNARY_OP:
             return self.evalTernaryOp(sexp, env)
        elif code is Syntax.AMBI1_OP:
            return self.evalAmbi1Op(sexp, env)
        elif code is Syntax.AMBI2_OP:
            return self.evalAmbi2Op(sexp, env)
        elif code is Syntax.N_ARY_OP:
            return self.evalNaryOp(sexp, env)
        elif code is Syntax.TRY:
            return self.evalTry(sexp, env)
        elif code is Syntax.FOR:
            return self.evalFor(sexp, env)
        else:
            raise PLambdaException("huh?")

    def evalPrimitiveDataOp(self, sexp, env):
        (a0, a1) = sexp.spine
        assert isinstance(a0, Atom)
        assert isinstance(a1, Atom)
        op = a0.string
        data = a1.string
        try:
            if op is SymbolTable.INT:
                return int(data)
            elif op is SymbolTable.FLOAT:
                return float(data)
            else:
                return True if data.lower() == 'true' else False
        except Exception as e:
            sys.stderr.write('evalPrimitiveDataOp: {0} {1}\n'.format(str(e), a0.location))
            return 0 if op is not SymbolTable.FLOAT else 0.0


    def importmod(self, val):
        try:
            
            if isString(val):
                module = importlib.import_module(val)
                if module is not None:
                    self.modules[val] = module
                    return True
                else:
                    sys.stderr.write('Module {0} not found'.format(val))
            return False

        except ImportError as ie:
            print(ie)
            return False


    def evalGlobal(self, val, loc):
        return pythonGlobals.get(val, None)

    def callUnaryOp(self, op, val, location):
        """Calls the unary op with argument val.

           It returns (True, op(val) if everything is OK, otherwise 
        it returns (False, Exception).
        """
        retval = None
        try:
            if  op is SymbolTable.LOAD:
                if isString(val):
                    retval = self.load(val)
                else:
                    retval = False
            elif op is SymbolTable.IMPORT:
                retval = self.importmod(val)
            elif op is SymbolTable.ISNONE:
                retval = val is None
            elif op is SymbolTable.ISINT:
                retval = isinstance(val, int)
            elif op is SymbolTable.ISFLOAT:
                retval = isinstance(val, float)
            elif op is SymbolTable.ISOBJECT:
                retval = val is not None
            elif op is SymbolTable.GLOBAL:
                retval = self.evalGlobal(val, location)
            elif op is SymbolTable.THROW:
                return (False,  val)
            elif op is SymbolTable.FETCH:
                if val in self.uid2object:
                    retval = self.uid2object[val]
                else:
                    retval = None
            elif op is SymbolTable.GETUID:
                if val in self.object2uid:
                    retval = self.object2uid[val]
                else:
                    retval = None
            elif op is SymbolTable.NOT:
                retval = True if val is False else False
            else:
                return (False, PlambdaException("Unrecognized unary op {0} {1}".format(op, location)))
        except Exception as e:
            return (False, PLambdaException('callUnaryOp {0} {1} threw {2}'.format(op, location, str(e))))

        return (True, retval)
        
    def evalUnaryOp(self, sexp, env):
        """Recursive version. Soon to be deprecated.
        """
        (uop, arg) =  sexp.spine
        assert isinstance(uop, Atom)
        op = uop.string
        val = self.eval(arg, env)

        (ok, retval) = self.callUnaryOp(op, val, sexp.location)

        if ok:
            return retval
        else:
            raise retval
        

    def load(self, filename):
        if isString(filename) and os.path.exists(filename):
            codelist = parseFromFile(filename)
            for c in codelist:
                self.evaluate(c)
            return True
        return False

    
    def showDefinitions(self, sb=None):
        """Either writes the definitions out to stderr, or the optional StringBuffer passed in.
        """
        func = sb.append if sb else sys.stderr.write
        for key, value in self.definitions.iteritems():
            func('{0}  -->  {1}\n'.format(key, value))
        return sb

    def showCode(self, sb=None):
        """Either writes the code out to stderr, or the optional StringBuffer passed in.
        """
        func = sb.append if sb else sys.stderr.write
        for key, value in self.code.iteritems():
            func('{0}  -->  {1}\n'.format(key, value))
        return sb

    def showUIDs(self, sb=None):
        """Either writes the UIDs out to stderr, or the optional StringBuffer passed in.
        """
        func = sb.append if sb else sys.stderr.write
        for key, value in self.uid2object.iteritems():
            func('{0}  -->  {1}\n'.format(key, value))
        return sb
    
            





from plambda.antlr4.PLambdaVisitor import PLambdaVisitor
from plambda.eval.SymbolTable import SymbolTable
from plambda.visitor.ParseError import ParseError
from plambda.eval.Code import SExpression, Atom, StringLiteral, Syntax, Location
from plambda.util.StringBuffer import StringBuffer

class Visitor(PLambdaVisitor):

    
    def __init__(self, filename):
        self.filename = filename

    # Visit a parse tree produced by PLambdaParser#unit.
    def visitUnit(self, ctx):
        retval = []
        for exp in ctx.expression():
            retval.append(self.visit(exp))
        return retval

    # Visit a parse tree produced by PLambdaParser#stringLiteral.
    def visitStringLiteral(self, ctx):
        t = ctx.STRING().getSymbol()
        location = Location(self.filename, t.line)
        return StringLiteral(deslashify(t.text), location)

    # Visit a parse tree produced by PLambdaParser#identifierLiteral.
    def visitIdentifierLiteral(self, ctx):
        t = ctx.ID().getSymbol()
        location = Location(self.filename, t.line)
        return  Atom(t.text, location)

    def visitExpressionList(self, code, tsym, explist):
        t = tsym.getSymbol()
        location = Location(self.filename, t.line)
        retval = [Atom(SymbolTable.canonicalize(t.text), location)]
        for e in explist:
            retval.append(self.visit(e))
        return SExpression(code, tuple(retval), location)

    def visitImplicitSeq(self, explist):
        retval = []
        location = None
        for e in explist:
            val = self.visit(e)
            if location is None:
                location = val.location
            retval.append(val)
        #if there is an implicit seq, make it explicit    
        if len(retval) > 1:
            retval.insert(0, Atom(SymbolTable.SEQ, location))
            return SExpression(Syntax.SEQ, tuple(retval), location)
        else:
            return retval[0]
            
    # Visit a parse tree produced by PLambdaParser#invokeExpression.
    def visitInvokeExpression(self, ctx):
        return self.visitExpressionList(Syntax.INVOKE, ctx.INVOKE(), ctx.expression())
    
    # Visit a parse tree produced by PLambdaParser#seqExpression.
    def visitSeqExpression(self, ctx):
        return self.visitExpressionList(Syntax.SEQ, ctx.SEQ(), ctx.expression())

    # Visit a parse tree produced by PLambdaParser#applyExpression.
    def visitApplyExpression(self, ctx):
        return self.visitExpressionList(Syntax.APPLY, ctx.APPLY(), ctx.expression())

    # Visit a parse tree produced by PLambdaParser#unaryExpression.
    def visitUnaryExpression(self, ctx):
        t = ctx.UNARY_OP().getSymbol()
        location = Location(self.filename, t.line)
        rawop = t.text
        canonicalop = SymbolTable.canonicalize(rawop)
        if canonicalop is not None:
            op = Atom(canonicalop, location)
            return SExpression(Syntax.UNARY_OP, (op, self.visit(ctx.expression())), location)
        else:
            raise ParseError('Ooopsy: {0} not found in the SymbolTable'.format(rawop))

    # Visit a parse tree produced by PLambdaParser#binaryExpression.
    def visitBinaryExpression(self, ctx):
        return self.visitExpressionList(Syntax.BINARY_OP, ctx.BINARY_OP(), ctx.expression())

    # Visit a parse tree produced by PLambdaParser#ternaryExpression.
    def visitTernaryExpression(self, ctx):
        return self.visitExpressionList(Syntax.TERNARY_OP, ctx.TERNARY_OP(), ctx.expression())

    # Visit a parse tree produced by PLambdaParser#lambdaExpression.
    def visitLambdaExpression(self, ctx):
        lineno = ctx.LAMBDA().getSymbol().line
        location = Location(self.filename, lineno)
        params = self.visitParameterList(ctx.parameterList())
        body = self.visitImplicitSeq(ctx.expression())
        lamb = Atom(SymbolTable.LAMBDA, location) 
        return SExpression(Syntax.LAMBDA, (lamb, params, body), location)

    # Visit a parse tree produced by PLambdaParser#parameterList.
    def visitParameterList(self, ctx):
        retval = []
        lineno = None
        for p in ctx.parameter():
            if lineno is None:
                lineno = p.ID().getSymbol().line
            retval.append(self.visitParameter(p))
        return SExpression(None, tuple(retval), Location(self.filename, lineno))

    # Visit a parse tree produced by PLambdaParser#parameter.
    def visitParameter(self, ctx):
        t = ctx.ID().getSymbol()
        location = Location(self.filename, t.line)
        return Atom(t.text, location)
            
    # Visit a parse tree produced by PLambdaParser#letExpression.
    def visitLetExpression(self, ctx):
        lineno = ctx.LET().getSymbol().line
        location = Location(self.filename, lineno)
        bindings = self.visitBindingList(ctx.bindingList())
        body = self.visitImplicitSeq(ctx.expression())
        let = Atom(SymbolTable.LET, location) 
        return SExpression(Syntax.LET, (let, bindings, body), location)
    
    # Visit a parse tree produced by PLambdaParser#bindingList.
    def visitBindingList(self, ctx):
        retval = []
        location = None
        for b in ctx.bindingPair():
            bp = self.visitBindingPair(b)
            if location is None:
                location = bp.location
            retval.append(bp)
        return SExpression(None, tuple(retval), location)

    # Visit a parse tree produced by PLambdaParser#binding_pair.
    def visitBindingPair(self, ctx):
        lineno = ctx.parameter().ID().getSymbol().line
        location = Location(self.filename, lineno)
        return SExpression(None,
                           (self.visitParameter(ctx.parameter()), self.visit(ctx.expression())),
                           location)

    # Visit a parse tree produced by PLambdaParser#defineExpression.
    def visitDefineExpression(self, ctx):
        lineno = ctx.DEFINE().getSymbol().line
        location = Location(self.filename, lineno)
        define = Atom(SymbolTable.DEFINE, location)
        paramList = ctx.parameterList()
        params = None
        if paramList is not None:
            params = self.visitParameterList(paramList)
        body = self.visitImplicitSeq(ctx.expression())
        id = Atom(intern(str(ctx.ID().getSymbol().text)), Location(self.filename, ctx.ID().getSymbol().line))
        if params is not None:
            return SExpression(Syntax.DEFINE, (define, id, params, body), location)
        else:
            return SExpression(Syntax.DEFINE, (define, id, body), location)


    # Visit a parse tree produced by PLambdaParser#forExpression.
    def visitForExpression(self, ctx):
        lineno = ctx.FOR().getSymbol().line
        location = Location(self.filename, lineno)
        f = Atom(SymbolTable.FOR, location)
        t = ctx.ID().getSymbol()
        location = Location(self.filename, t.line)
        id = Atom(t.text, location)
        return SExpression(Syntax.FOR, (f, id,
                                        self.visit(ctx.rangeExpression()),
                                        self.visitImplicitSeq(ctx.expression())),
                           location)


    # Visit a parse tree produced by PLambdaParser#tryExpression.
    def visitTryExpression(self, ctx):
        lineno = ctx.TRY().getSymbol().line
        location = Location(self.filename, lineno)
        tr = Atom(SymbolTable.TRY, location)
        return SExpression(Syntax.TRY, (tr,
                                        self.visitImplicitSeq(ctx.expression()),
                                        self.visitCatchExpression(ctx.catchExpression())),
                           location)
    
    # Visit a parse tree produced by PLambdaParser#catchExpression.
    def visitCatchExpression(self, ctx):
        lineno = ctx.CATCH().getSymbol().line
        location = Location(self.filename, lineno)
        ctch = Atom(SymbolTable.CATCH, location)
        return SExpression(Syntax.CATCH, (ctch,
                     self.visitParameter(ctx.parameter()),
                     self.visitImplicitSeq(ctx.expression())),
                    location)
                    

    # Visit a parse tree produced by PLambdaParser#dataExpression.
    def visitDataExpression(self, ctx):
        t = ctx.PRIMITIVE_DATA_OP().getSymbol()
        lineno = t.line
        location = Location(self.filename, t.line)
        op = Atom(SymbolTable.canonicalize(t.text), location)
        return SExpression(Syntax.PRIMITIVE_DATA_OP, (op, self.visitData(ctx.data())), location)

    # Visit a parse tree produced by PLambdaParser#data.
    def visitData(self, ctx):
        data = None
        if ctx.ID() != None:
            data = ctx.ID()
        elif ctx.NUMBER() != None:
            data = ctx.NUMBER()
        else:
            data = ctx.CHARACTER()
        return Atom(data.getSymbol().text, Location(self.filename, data.getSymbol().line))

    
    # Visit a parse tree produced by PLambdaParser#oneOrMoreExpression.
    def visitOneOrMoreExpression(self, ctx):
        return self.visitExpressionList(Syntax.AMBI1_OP, ctx.AMBI1_OP(), ctx.expression());

    # Visit a parse tree produced by PLambdaParser#twoOrMoreExpression.
    def visitTwoOrMoreExpression(self, ctx):
        return self.visitExpressionList(Syntax.AMBI2_OP, ctx.AMBI2_OP(), ctx.expression());

    # Visit a parse tree produced by PLambdaParser#naryExpression.
    def visitNaryExpression(self, ctx):
        return self.visitExpressionList(Syntax.N_ARY_OP, ctx.N_ARY_OP(), ctx.expression()); 


def deslashify(string):
    return deslashify_aux(string[0], string)
    

def deslashify_aux(dl, string):
    """Interprets the slashes in a raw string.
    """
    sb = StringBuffer()
    slash = False
    lstr = len(string)
    i = 0;
    #we'll need to be more careful if we are going to allow 'strings like this'
    assert(string[0] == dl and string[-1] == dl)
    for c in string:
        i += 1 
        if slash:
            slash = False
            if c == 'n':
                sb.append('\n')
            elif c == dl:
                #cannot end with a dangling slash
                if i == lstr:
                    raise ParseError('Illegal dangling \\ in String')
                sb.append(dl)
            elif c == 't':
                sb.append('\t')
            elif c == 'f':
                sb.append('\f')
            elif c == 'b':
                sb.append('\b')
            elif c == '\\':
                sb.append('\\')
            else:
                raise ParseError('Illegal escape character in String: {0}'.format(c))
        elif c != '\\':
            sb.append(c)
            assert(slash == False)
        else:
            slash = True
            continue

    return str(sb)



import sys

class SymbolTable(object):

    _symbol_table = {
        intern('and'):           intern('and'),
        intern('apply'):         intern('apply'),
        intern('array'):         intern('array'),
        intern('aset'):          intern('aset'),
        intern('boolean'):       intern('boolean'),
        intern('byte'):          intern('byte'),
        intern('catch'):         intern('catch'),
        intern('char'):          intern('char'),
        intern('concat'):        intern('concat'),
        intern('define'):        intern('define'),
        intern('/'):             intern('/'),
        intern('do'):            intern('do'),
        intern('double'):        intern('double'),
        intern('=='):            intern('=='),
        intern('is'):            intern('is'),
        intern('fetch'):         intern('fetch'),
        intern('float'):         intern('float'),
        intern('for'):           intern('for'),
        intern('<='):            intern('<='),
        intern('get'):           intern('get'),
        intern('getattr'):       intern('getattr'),
        intern('getuid'):        intern('getuid'),
        intern('global'):        intern('global'),
        intern('<'):             intern('<'),
        intern('if'):            intern('if'),
        intern('import'):        intern('import'),
        intern('instanceof'):    intern('instanceof'),
        intern('int'):           intern('int'),
        intern('invoke'):        intern('invoke'),
        intern('isnone'):        intern('isnone'),
        intern('isfloat'):       intern('isfloat'),
        intern('isint'):         intern('isint'),
        intern('isobject'):      intern('isobject'),
        intern('kwapply'):       intern('kwapply'),
        intern('lambda'):        intern('lambda'),
        intern('>='):            intern('>='),
        intern('let'):           intern('let'),
        intern('load'):          intern('load'),
        intern('long'):          intern('long'),
        intern('lookup'):        intern('lookup'),
        intern('>'):             intern('>'),
        intern('-'):             intern('-'),
        intern('mktuple'):       intern('mktuple'),
        intern('mklist'):        intern('mklist'),
        intern('mkdict'):        intern('mkdict'),
        intern('modify'):        intern('modify'),
        intern('%'):             intern('%'),
        intern('narrow'):        intern('narrow'),
        intern('!='):            intern('!='),
        intern('not'):           intern('not'),
        intern('none'):          intern('none'),
        intern('object'):        intern('object'),
        intern('or'):            intern('or'),
        intern('+'):             intern('+'),
        intern('seq'):           intern('seq'),
        intern('setattr'):       intern('setattr'),
        intern('setuid'):        intern('setuid'),
        intern('short'):         intern('short'),
        intern('sinvoke'):       intern('sinvoke'),
        intern('supdate'):       intern('supdate'),
        intern('throw'):         intern('throw'),
        intern('*'):             intern('*'),
        intern('try'):           intern('try'),
        intern('update'):        intern('update'),
        }

        
    AND         = intern('and')
    APPLY       = intern('apply')
    ARRAY       = intern('array')
    ASET        = intern('aset')
    BOOLEAN     = intern('boolean')
    BYTE        = intern('byte')
    CATCH       = intern('catch')
    CHAR        = intern('char')
    CONCAT      = intern('concat')
    DEFINE      = intern('define')
    DIVIDE      = intern('/')
    DO          = intern('do')
    DOUBLE      = intern('double')
    IS          = intern('is')
    EQUALS      = intern('==')
    FETCH       = intern('fetch')
    FLOAT       = intern('float')
    FOR         = intern('for')
    GEQ         = intern('<=')
    GET         = intern('get')
    GETATTR     = intern('getattr')
    GETUID      = intern('getuid')
    GLOBAL      = intern('global')
    GT          = intern('<')
    IF          = intern('if')
    IMPORT      = intern('import')
    INSTANCEOF  = intern('instanceof')
    INT         = intern('int')
    INVOKE      = intern('invoke')
    ISNONE      = intern('isnone')
    ISFLOAT     = intern('isfloat')
    ISINT       = intern('isint')
    ISOBJECT    = intern('isobject')
    KWAPPLY     = intern('kwapply')
    LAMBDA      = intern('lambda')
    LEQ         = intern('>=')
    LET         = intern('let')
    LOAD        = intern('load')
    LONG        = intern('long')
    LOOKUP      = intern('lookup')
    LT          = intern('>')
    MINUS       = intern('-')
    MKTUPLE     = intern('mktuple')
    MKLIST      = intern('mklist')
    MKDICT      = intern('mkdict')
    MODULO      = intern('%')
    MODIFY      = intern('modify')
    NARROW      = intern('narrow')
    NEQ         = intern('!=')
    NOT         = intern('not')
    NONE        = intern('none')
    OBJECT      = intern('object')
    OR          = intern('or')
    PLUS        = intern('+')
    SEQ         = intern('seq')
    SETATTR     = intern('setattr')
    SETUID      = intern('setuid')
    SHORT       = intern('short')
    SINVOKE     = intern('sinvoke')
    SUPDATE     = intern('supdate')
    THROW       = intern('throw')
    TIMES       = intern('*')
    TRY         = intern('try')
    UPDATE      = intern('update')
    
    @staticmethod
    def canonicalize(name):
        return SymbolTable._symbol_table.get(name.lower())
    



import os, platform, sys,  select, time, traceback

from ..util.StringBuffer import StringBuffer

from ..visitor.Parser import parseFromString
from ..eval.Interpreter  import Interpreter
from ..eval.PLambdaException import PLambdaException
from ..version import plambda_version

def main():
    rep(sys.argv[1] if len(sys.argv) == 2 else None)
    return 0


def snarf(fp, delay):
    """ read as much as possible without blocking. (won't work on windows) """
    if platform.system == 'Windows':
        return sys.stdin.readline().strip()
    else:
        sb = StringBuffer()
        count = 0
        while True:
            while sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
                line = sys.stdin.readline()
                if line:
                    count += 1
                    sb.append(line)
            else:
                if sb.isempty():
                    time.sleep(delay)
                else:
                    if count > 1:
                        return str(sb)
                    else:
                        return str(sb).strip()



def rep(filename):
    """The Read Eval Print loop for the plambda language.
    """
    interpreter = Interpreter()

    debug = False
    
    try:

        interpreter.load(filename)
        
        sys.stdout.write(WELCOME.format(plambda_version))

        while True:
            try:
                sys.stdout.write('> ')
                sys.stdout.flush()
                line = snarf(sys.stdin, 0.2)
                if line == 'q':
                    return 0
                elif line == 'v':
                    debug = not debug
                elif line == '?':
                    sys.stdout.write(INSTRUCTIONS)
                elif line == 'd':
                    interpreter.showDefinitions()
                elif line == 'u':
                    interpreter.showUIDs()
                elif line == 's':
                    interpreter.showCode()
                else:
                    if line:
                        if debug:
                            print 'rep: line  = ', line
                        code = parseFromString(line)
                        for c in code:
                            if c is not None:
                                if debug:
                                    print 'rep: sexp  = ', c
                                value = interpreter.evaluate(c)
                                if debug:
                                    print 'rep: value = ', value
                                print value
            except PLambdaException as e:
                print 'PLambda.rep PLambdaException: ', e
            except Exception as e:
                print 'PLambda.rep Exception: ', e
                traceback.print_exc(file=sys.stderr)
                
    except KeyboardInterrupt:
        return 0



WELCOME = """
Welcome to the PLambda interface to Python (version {0}), type ? for help.
"""

INSTRUCTIONS="""
Type one of the following:
\tany valid plambda expression to be evaluated, or
\tq to quit  (or quit)
\t? to see these instructions
\td to see the current definitions
\ts <name> to see the raw definition of <name>
\tu to see the current uids
\tv to toggle the degree of verbosity in error reports
"""
    



import os, sys
from antlr4 import *

from ..antlr4.PLambdaLexer import PLambdaLexer

from ..antlr4.PLambdaParser import PLambdaParser

from .Visitor import Visitor


def main():
    if len(sys.argv) != 2:
        print 'Usage: {0} <plambda file>'.format(sys.argv[0])
    else:
        codelist = parseFromFile(sys.argv[1])
        for c in codelist:
            print str(c)
            print repr(c)
        return 0

import sys

def parseFromFile(filename):
    return parseFromStream(FileStream(filename), filename)
 
def parseFromString(string):
    return parseFromStream(InputStream(string), "stdin")

def parseFromStream(stream, source):
    lexer = PLambdaLexer(stream)
    stream = CommonTokenStream(lexer)
    parser = PLambdaParser(stream)
    tree = parser.unit()
    visitor = Visitor(source)
    return visitor.visit(tree)
    



import sys
import inspect

from .actorlib import send, receive

class Actor(object):

    _debug = False
    
    def __init__(self, name, client):
        """Creates an Actor object with the given name and client."""
        self.client = client
        self.name = name
        # should probably complain if client is not an object...
        self.methods = inspect.getmembers(self.client, callable)
        debug('methods: {0}\n'.format(self.methods))


    def begin(self):
        """Receives and processes messages in an infinite loop. """
        while True:
            incoming =  receive()
            if incoming is None:
                return 0
     	    (sender, msg) = incoming
            send(sender, self.name, self.process_message(sender, msg))


    def process_message(self, sender, msg):
        """The beginnings of a REP for imaude messages. 

        Currently processes the message as:

         verb arg0 arg1 ... argN
        
        and attempts to invoke the 'verb' method on the 'client' object
        with the remaining arguments. I.e. 

        client.verb(arg0, arg1, ..., argN)

        If this is kosher it returns the string representation of 
        the resulting 'client' object.
        """

        #split the args into verb and rest
        arglist = msg.split()
        verb = arglist[0]
        args = arglist[1:]

        debug('{0} got message with verb = {1} and args = {2}\n'.format(self.name, verb, args))
                
        # see if we can call the 'verb' on the 'client' with the given 'args'
        success = False
        for (name, value) in self.methods:
            if name == verb:
                argspec = inspect.getargspec(value)
                debug('argspec: {0}'.format(argspec))
                success = True
                if len(argspec.args) - 1  == len(args): # do not count 'self'
                    value(*args)
                else:
                    fmsg = 'Actor arity of msg {0} does not match argsec: {1}'
                    emsg = fmsg.format(args, argspec.args[1:])
                    sys.stderr.write(emsg)
                    success = False
                break

        # return the result to the sender
        if success:
            return str(self.client)
        else:
            return 'unknown command'


            
def debug(emsg):
    """If the Actor.flag is set to True, then it writes the msg to stderr."""
    if Actor._debug:
        sys.stderr.write(emsg)
            




from ..util.StringBuffer import StringBuffer

#may as well retain information gleaned from the parser
#since it will help in the dispatching
class Syntax(object):

    SEQ               = 0
    LET               = 1
    DEFINE            = 2
    LAMBDA            = 3
    INVOKE            = 4
    APPLY             = 5
    PRIMITIVE_DATA_OP = 6
    UNARY_OP          = 7
    BINARY_OP         = 8
    TERNARY_OP        = 9
    AMBI1_OP          = 10
    AMBI2_OP          = 11
    N_ARY_OP          = 12
    TRY               = 13
    FOR               = 14
    QUOTE             = 15
    CATCH             = 16


class Location(object):

    def __init__(self, filename, lineno):
        self.filename = filename
        self.lineno = lineno


    def __str__(self):
        return '@{0}:{1}'.format(self.filename, self.lineno)



class SExpression(object):

    def __init__(self, code, spine, location):
        self.code = code
        self.spine = spine
        self.location = location
        self.string = self.toString()


    # repr's goal is to be unambiguous
    def __repr__(self):
        return '{0}{1}'.format(self.string, self.location)

    # str's goal is to be readable
    def __str__(self):
        return self.string


    def toString(self):
        sb = StringBuffer()
        first = True
        sb.append('(')
        for c in self.spine:
            if first:
                first = False
            else:
                sb.append(' ')
            sb.append(str(c))
        sb.append(')')
        return str(sb)
    


class Atom(object):

    def __init__(self, uni, location):
        self.string = intern(str(uni))
        self.location = location

    # repr's goal is to be unambiguous
    def __repr__(self):
        return '{0}{1}'.format(self.string, self.location)

    # str's goal is to be readable
    def __str__(self):
        return self.string

class StringLiteral(object):

    def __init__(self, uni, location):
        self.string = intern(str(uni)[1:-1]) # remove the quotes
        self.location = location

    # repr's goal is to be unambiguous
    def __repr__(self):
        return '{0}{1}'.format(self.string, self.location)

    # str's goal is to be readable
    def __str__(self):
        return self.string




# Generated from java-escape by ANTLR 4.5
from antlr4 import *

# This class defines a complete generic visitor for a parse tree produced by PLambdaParser.

class PLambdaVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by PLambdaParser#unit.
    def visitUnit(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#seqExpression.
    def visitSeqExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#letExpression.
    def visitLetExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#defineExpression.
    def visitDefineExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#lambdaExpression.
    def visitLambdaExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#invokeExpression.
    def visitInvokeExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#applyExpression.
    def visitApplyExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#dataExpression.
    def visitDataExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#unaryExpression.
    def visitUnaryExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#binaryExpression.
    def visitBinaryExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#ternaryExpression.
    def visitTernaryExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#oneOrMoreExpression.
    def visitOneOrMoreExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#twoOrMoreExpression.
    def visitTwoOrMoreExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#naryExpression.
    def visitNaryExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#tryExpression.
    def visitTryExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#forExpression.
    def visitForExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#stringLiteral.
    def visitStringLiteral(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#identifierLiteral.
    def visitIdentifierLiteral(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#noneLiteral.
    def visitNoneLiteral(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#parameterList.
    def visitParameterList(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#parameter.
    def visitParameter(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#bindingList.
    def visitBindingList(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#bindingPair.
    def visitBindingPair(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#catchExpression.
    def visitCatchExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#rangeExpression.
    def visitRangeExpression(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#data.
    def visitData(self, ctx):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PLambdaParser#token.
    def visitToken(self, ctx):
        return self.visitChildren(ctx)





# from distutils.core import setup
from setuptools import setup, find_packages
import os
import glob

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# use the in house version number so we stay in synch with ourselves.
from plambda.version import plambda_version
    
setup(
    name='PLambda',
    version=plambda_version, 
    description='The PLambda language',
    long_description=long_description,
    url='https://github.com/SRI-CSL/PLambda',
    author='Ian A. Mason',
    author_email='iam@csl.sri.com',
    

    packages=find_packages(exclude=['tests']),
    
    entry_points = {
        'console_scripts': [
            'plambda = plambda.eval.PLambda:main',
            'pyactor = plambda.actors.pyactor:main',
            'plambdaparse = plambda.visitor.Parser:main',
        ],
    },


    
    license='MIT',
    
    install_requires=[
        "psutil >= 4.3.0",
        "antlr4-python2-runtime >= 4.5.3"
    ],
    
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'Topic :: System :: Distributed Computing',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
)



from .Code import Atom, Location

from .Environment import Environment

class Closure(object):

    def __init__(self, interpreter, params, body, env, location):

        self.interpreter = interpreter
        self.params = params
        self.arity = len(params.spine)
        self.body = body
        self.env = env
        self.location = location
        
    def applyClosure(self, *args):
        nenv = Environment(self.env)
        for (p, v) in zip(self.params.spine, args):
            #print "extending: {0} --> {1}".format(p, v)
            nenv.extend(p, v)
        return  self.interpreter.eval(self.body, nenv)
    

    
